@@ -1,6 +1,22 @@
+/*
+ * Copyright 2021 l1npengtul <l1npengtul@protonmail.com> / The Nokhwa Contributors
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
 use crate::{
-    mjpeg_to_rgb888, yuyv422_to_rgb888, CameraFormat, CameraInfo, CaptureAPIBackend,
-    CaptureBackendTrait, FrameFormat, NokhwaError, Resolution,
+    mjpeg_to_rgb888, yuyv422_to_rgb888, CameraControl, CameraFormat, CameraInfo, CaptureAPIBackend,
+    CaptureBackendTrait, FrameFormat, KnownCameraControls, NokhwaError, Resolution,
 };
 use flume::Receiver;
 use glib::Quark;
@@ -15,7 +31,8 @@ use gstreamer_app::{AppSink, AppSinkCallbacks};
 use gstreamer_video::{VideoFormat, VideoInfo};
 use image::{ImageBuffer, Rgb};
 use regex::Regex;
-use std::{collections::HashMap, str::FromStr};
+use std::any::Any;
+use std::{borrow::Cow, collections::HashMap, str::FromStr};
 
 type PipelineGenRet = (Element, AppSink, Receiver<ImageBuffer<Rgb<u8>, Vec<u8>>>);
 
@@ -23,6 +40,7 @@ type PipelineGenRet = (Element, AppSink, Receiver<ImageBuffer<Rgb<u8>, Vec<u8>>>
 /// To see what this does, please see [`CaptureBackendTrait`].
 /// # Quirks
 /// - `Drop`-ing this may cause a `panic`.
+/// - Setting controls is not supported.
 pub struct GStreamerCaptureDevice {
     pipeline: Element,
     app_sink: AppSink,
@@ -47,10 +65,10 @@ impl GStreamerCaptureDevice {
         };
 
         if let Err(why) = gstreamer::init() {
-            return Err(NokhwaError::CouldntOpenDevice(format!(
-                "Failed to initialize GStreamer: {}",
-                why.to_string()
-            )));
+            return Err(NokhwaError::InitializeError {
+                backend: CaptureAPIBackend::GStreamer,
+                error: why.to_string(),
+            });
         }
 
         let (camera_info, caps) = {
@@ -64,28 +82,29 @@ impl GStreamerCaptureDevice {
                     )))
                 }
             };
-            let _video_filter_id = match device_monitor
-                .add_filter(Some("Video/Source"), Some(&video_caps))
-            {
-                Some(id) => id,
-                None => return Err(NokhwaError::CouldntOpenDevice(
-                    "Failed to generate Device Monitor Filter ID with video/x-raw and Video/Source"
-                        .to_string(),
-                )),
-            };
+            let _video_filter_id =
+                match device_monitor.add_filter(Some("Video/Source"), Some(&video_caps)) {
+                    Some(id) => id,
+                    None => {
+                        return Err(NokhwaError::StructureError {
+                            structure: "Video Filter ID Video/Source".to_string(),
+                            error: "Null".to_string(),
+                        })
+                    }
+                };
             if let Err(why) = device_monitor.start() {
-                return Err(NokhwaError::CouldntOpenDevice(format!(
-                    "Failed to start device monitor: {}",
-                    why.to_string()
-                )));
+                return Err(NokhwaError::StructureError {
+                    structure: "Device Monitor".to_string(),
+                    error: format!("Not started, {}", why),
+                });
             }
             let device = match device_monitor.devices().get(index) {
                 Some(dev) => dev.clone(),
                 None => {
-                    return Err(NokhwaError::CouldntOpenDevice(format!(
-                        "Failed to find device at index {}",
-                        index
-                    )))
+                    return Err(NokhwaError::OpenDeviceError(
+                        index.to_string(),
+                        "No device".to_string(),
+                    ))
                 }
             };
             device_monitor.stop();
@@ -125,8 +144,12 @@ impl GStreamerCaptureDevice {
 }
 
 impl CaptureBackendTrait for GStreamerCaptureDevice {
-    fn camera_info(&self) -> CameraInfo {
-        self.camera_info.clone()
+    fn backend(&self) -> CaptureAPIBackend {
+        CaptureAPIBackend::GStreamer
+    }
+
+    fn camera_info(&self) -> &CameraInfo {
+        &self.camera_info
     }
 
     fn camera_format(&self) -> CameraFormat {
@@ -139,7 +162,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
             self.stop_stream()?;
             reopen = true;
         }
-        let (pipeline, app_sink, receiver) = generate_pipeline(new_fmt, *self.camera_info.index())?;
+        let (pipeline, app_sink, receiver) = generate_pipeline(new_fmt, self.camera_info.index())?;
         self.pipeline = pipeline;
         self.app_sink = app_sink;
         self.receiver = receiver;
@@ -153,7 +176,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
     #[allow(clippy::too_many_lines)]
     #[allow(clippy::cast_sign_loss)]
     fn compatible_list_by_resolution(
-        &self,
+        &mut self,
         fourcc: FrameFormat,
     ) -> Result<HashMap<Resolution, Vec<u32>>, NokhwaError> {
         let mut resolution_map = HashMap::new();
@@ -171,7 +194,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                 let width = match capability.get::<i32>("width") {
                                     Ok(w) => w,
                                     Err(why) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Capibilities by Resolution: Width"
                                                 .to_string(),
                                             error: why.to_string(),
@@ -181,7 +204,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                 let height = match capability.get::<i32>("height") {
                                     Ok(w) => w,
                                     Err(why) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Capibilities by Resolution: Height"
                                                 .to_string(),
                                             error: why.to_string(),
@@ -196,7 +219,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             format!("{:?}", s)
                                         }
                                         Err(why) => {
-                                            return Err(NokhwaError::CouldntQueryDevice {
+                                            return Err(NokhwaError::GetPropertyError {
                                                 property: "Framerates".to_string(),
                                                 error: format!(
                                                     "Failed to make framerates into string: {}",
@@ -206,7 +229,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                         }
                                     },
                                     Err(_) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Framerates".to_string(),
                                             error: "Failed to get framerates: doesnt exist!"
                                                 .to_string(),
@@ -218,7 +241,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                     let fraction_string: Vec<&str> =
                                         m.as_str().split('/').collect();
                                     if fraction_string.len() != 2 {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: format!("Fraction framerate had more than one demoninator: {:?}", fraction_string) });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: format!("Fraction framerate had more than one demoninator: {:?}", fraction_string) });
                                     }
 
                                     if let Some(v) = fraction_string.get(1) {
@@ -226,14 +249,14 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             continue; // swallow error
                                         }
                                     } else {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: "No framerate denominator? Shouldn't happen, please report!".to_string() });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: "No framerate denominator? Shouldn't happen, please report!".to_string() });
                                     }
 
                                     if let Some(numerator) = fraction_string.get(0) {
                                         match numerator.parse::<u32>() {
                                             Ok(fps) => fps_vec.push(fps),
                                             Err(why) => {
-                                                return Err(NokhwaError::CouldntQueryDevice {
+                                                return Err(NokhwaError::GetPropertyError {
                                                     property: "Framerates".to_string(),
                                                     error: format!(
                                                         "Failed to parse numerator: {}",
@@ -243,7 +266,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             }
                                         }
                                     } else {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: "No framerate numerator? Shouldn't happen, please report!".to_string() });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: "No framerate numerator? Shouldn't happen, please report!".to_string() });
                                     }
                                 }
                                 resolution_map
@@ -259,7 +282,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                 let width = match capability.get::<i32>("width") {
                                     Ok(w) => w,
                                     Err(why) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Capibilities by Resolution: Width"
                                                 .to_string(),
                                             error: why.to_string(),
@@ -269,7 +292,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                 let height = match capability.get::<i32>("height") {
                                     Ok(w) => w,
                                     Err(why) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Capibilities by Resolution: Height"
                                                 .to_string(),
                                             error: why.to_string(),
@@ -284,7 +307,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             format!("{:?}", s)
                                         }
                                         Err(why) => {
-                                            return Err(NokhwaError::CouldntQueryDevice {
+                                            return Err(NokhwaError::GetPropertyError {
                                                 property: "Framerates".to_string(),
                                                 error: format!(
                                                     "Failed to make framerates into string: {}",
@@ -294,7 +317,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                         }
                                     },
                                     Err(_) => {
-                                        return Err(NokhwaError::CouldntQueryDevice {
+                                        return Err(NokhwaError::GetPropertyError {
                                             property: "Framerates".to_string(),
                                             error: "Failed to get framerates: doesnt exist!"
                                                 .to_string(),
@@ -306,7 +329,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                     let fraction_string: Vec<&str> =
                                         m.as_str().split('/').collect();
                                     if fraction_string.len() != 2 {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: format!("Fraction framerate had more than one demoninator: {:?}", fraction_string) });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: format!("Fraction framerate had more than one demoninator: {:?}", fraction_string) });
                                     }
 
                                     if let Some(v) = fraction_string.get(1) {
@@ -314,14 +337,14 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             continue; // swallow error
                                         }
                                     } else {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: "No framerate denominator? Shouldn't happen, please report!".to_string() });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: "No framerate denominator? Shouldn't happen, please report!".to_string() });
                                     }
 
                                     if let Some(numerator) = fraction_string.get(0) {
                                         match numerator.parse::<u32>() {
                                             Ok(fps) => fps_vec.push(fps),
                                             Err(why) => {
-                                                return Err(NokhwaError::CouldntQueryDevice {
+                                                return Err(NokhwaError::GetPropertyError {
                                                     property: "Framerates".to_string(),
                                                     error: format!(
                                                         "Failed to parse numerator: {}",
@@ -331,7 +354,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                                             }
                                         }
                                     } else {
-                                        return Err(NokhwaError::CouldntQueryDevice { property: "Framerates".to_string(), error: "No framerate numerator? Shouldn't happen, please report!".to_string() });
+                                        return Err(NokhwaError::GetPropertyError { property: "Framerates".to_string(), error: "No framerate numerator? Shouldn't happen, please report!".to_string() });
                                     }
                                 }
                                 resolution_map
@@ -342,7 +365,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
                 }
             }
             None => {
-                return Err(NokhwaError::CouldntQueryDevice {
+                return Err(NokhwaError::GetPropertyError {
                     property: "Device Caps".to_string(),
                     error: "No device caps!".to_string(),
                 })
@@ -358,16 +381,16 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
             Some(c) => {
                 for capability in c.iter() {
                     if capability.name() == "image/jpeg" {
-                        format_vec.push(FrameFormat::MJPEG)
+                        format_vec.push(FrameFormat::MJPEG);
                     } else if capability.name() == "video/x-raw"
                         && capability.get::<String>("format").unwrap_or_default() == *"YUY2"
                     {
-                        format_vec.push(FrameFormat::YUYV)
+                        format_vec.push(FrameFormat::YUYV);
                     }
                 }
             }
             None => {
-                return Err(NokhwaError::CouldntQueryDevice {
+                return Err(NokhwaError::GetPropertyError {
                     property: "Device Caps".to_string(),
                     error: "No device caps!".to_string(),
                 })
@@ -389,12 +412,12 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
     }
 
     fn frame_rate(&self) -> u32 {
-        self.camera_format.framerate()
+        self.camera_format.frame_rate()
     }
 
     fn set_frame_rate(&mut self, new_fps: u32) -> Result<(), NokhwaError> {
         let mut new_fmt = self.camera_format;
-        new_fmt.set_framerate(new_fps);
+        new_fmt.set_frame_rate(new_fps);
         self.set_camera_format(new_fmt)
     }
 
@@ -403,14 +426,54 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
     }
 
     fn set_frame_format(&mut self, _fourcc: FrameFormat) -> Result<(), NokhwaError> {
-        Err(NokhwaError::UnsupportedOperation(
+        Err(NokhwaError::UnsupportedOperationError(
             CaptureAPIBackend::GStreamer,
         ))
     }
 
+    fn supported_camera_controls(&self) -> Result<Vec<KnownCameraControls>, NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
+    fn camera_control(&self, _control: KnownCameraControls) -> Result<CameraControl, NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
+    fn set_camera_control(&mut self, _control: CameraControl) -> Result<(), NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
+    fn raw_supported_camera_controls(&self) -> Result<Vec<Box<dyn Any>>, NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
+    fn raw_camera_control(&self, _control: &dyn Any) -> Result<Box<dyn Any>, NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
+    fn set_raw_camera_control(
+        &mut self,
+        _control: &dyn Any,
+        _value: &dyn Any,
+    ) -> Result<(), NokhwaError> {
+        Err(NokhwaError::NotImplementedError(
+            CaptureAPIBackend::GStreamer.to_string(),
+        ))
+    }
+
     fn open_stream(&mut self) -> Result<(), NokhwaError> {
         if let Err(why) = self.pipeline.set_state(State::Playing) {
-            return Err(NokhwaError::CouldntOpenStream(format!(
+            return Err(NokhwaError::OpenStreamError(format!(
                 "Failed to set appsink to playing: {}",
                 why.to_string()
             )));
@@ -435,15 +498,15 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
     }
 
     fn frame(&mut self) -> Result<ImageBuffer<Rgb<u8>, Vec<u8>>, NokhwaError> {
-        let image_data = self.frame_raw()?;
         let cam_fmt = self.camera_format;
+        let image_data = self.frame_raw()?;
         let imagebuf =
-            match ImageBuffer::from_vec(cam_fmt.width(), cam_fmt.height(), image_data) {
+            match ImageBuffer::from_vec(cam_fmt.width(), cam_fmt.height(), image_data.to_vec()) {
                 Some(buf) => {
                     let rgbbuf: ImageBuffer<Rgb<u8>, Vec<u8>> = buf;
                     rgbbuf
                 }
-                None => return Err(NokhwaError::CouldntCaptureFrame(
+                None => return Err(NokhwaError::ReadFrameError(
                     "Imagebuffer is not large enough! This is probably a bug, please report it!"
                         .to_string(),
                 )),
@@ -451,11 +514,11 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
         Ok(imagebuf)
     }
 
-    fn frame_raw(&mut self) -> Result<Vec<u8>, NokhwaError> {
+    fn frame_raw(&mut self) -> Result<Cow<[u8]>, NokhwaError> {
         let bus = match self.pipeline.bus() {
             Some(bus) => bus,
             None => {
-                return Err(NokhwaError::CouldntCaptureFrame(
+                return Err(NokhwaError::ReadFrameError(
                     "The pipeline has no bus!".to_string(),
                 ))
             }
@@ -464,12 +527,10 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
         if let Some(message) = bus.timed_pop(ClockTime::from_seconds(0)) {
             match message.view() {
                 MessageView::Eos(..) => {
-                    return Err(NokhwaError::CouldntCaptureFrame(
-                        "Stream is ended!".to_string(),
-                    ))
+                    return Err(NokhwaError::ReadFrameError("Stream is ended!".to_string()))
                 }
                 MessageView::Error(err) => {
-                    return Err(NokhwaError::CouldntCaptureFrame(format!(
+                    return Err(NokhwaError::ReadFrameError(format!(
                         "Bus error: {}",
                         err.error().to_string()
                     )));
@@ -479,9 +540,9 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
         }
 
         match self.receiver.recv() {
-            Ok(msg) => Ok(msg.to_vec()),
+            Ok(msg) => Ok(Cow::from(msg.to_vec())),
             Err(why) => {
-                return Err(NokhwaError::CouldntCaptureFrame(format!(
+                return Err(NokhwaError::ReadFrameError(format!(
                     "Receiver Error: {}",
                     why.to_string()
                 )));
@@ -491,7 +552,7 @@ impl CaptureBackendTrait for GStreamerCaptureDevice {
 
     fn stop_stream(&mut self) -> Result<(), NokhwaError> {
         if let Err(why) = self.pipeline.set_state(State::Null) {
-            return Err(NokhwaError::CouldntStopStream(format!(
+            return Err(NokhwaError::StreamShutdownError(format!(
                 "Could not change state: {}",
                 why.to_string()
             )));
@@ -510,10 +571,10 @@ impl Drop for GStreamerCaptureDevice {
 fn webcam_pipeline(device: &str, camera_format: CameraFormat) -> String {
     match camera_format.format() {
         FrameFormat::MJPEG => {
-            format!("autovideosrc location=/dev/video{} ! image/jpeg,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("autovideosrc location=/dev/video{} ! image/jpeg,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
         FrameFormat::YUYV => {
-            format!("autovideosrc location=/dev/video{} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("autovideosrc location=/dev/video{} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
     }
 }
@@ -522,10 +583,10 @@ fn webcam_pipeline(device: &str, camera_format: CameraFormat) -> String {
 fn webcam_pipeline(device: &str, camera_format: CameraFormat) -> String {
     match camera_format.format() {
         FrameFormat::MJPEG => {
-            format!("v4l2src device=/dev/video{} ! image/jpeg, width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("v4l2src device=/dev/video{} ! image/jpeg, width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
         FrameFormat::YUYV => {
-            format!("v4l2src device=/dev/video{} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("v4l2src device=/dev/video{} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
     }
 }
@@ -534,10 +595,10 @@ fn webcam_pipeline(device: &str, camera_format: CameraFormat) -> String {
 fn webcam_pipeline(device: &str, camera_format: CameraFormat) -> String {
     match camera_format.format() {
         FrameFormat::MJPEG => {
-            format!("ksvideosrc device_index={} ! image/jpeg, width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("ksvideosrc device_index={} ! image/jpeg, width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
         FrameFormat::YUYV => {
-            format!("ksvideosrc device_index={} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.framerate())
+            format!("ksvideosrc device_index={} ! video/x-raw,format=YUY2,width={},height={},framerate={}/1 ! appsink name=appsink async=false sync=false", device, camera_format.width(), camera_format.height(), camera_format.frame_rate())
         }
     }
 }
@@ -550,11 +611,14 @@ fn generate_pipeline(fmt: CameraFormat, index: usize) -> Result<PipelineGenRet,
         {
             Ok(p) => p,
             Err(why) => {
-                return Err(NokhwaError::CouldntOpenDevice(format!(
-                    "Failed to open pipeline with args {}: {}",
-                    webcam_pipeline(format!("{}", index).as_str(), fmt),
-                    why.to_string()
-                )))
+                return Err(NokhwaError::OpenDeviceError(
+                    index.to_string(),
+                    format!(
+                        "Failed to open pipeline with args {}: {}",
+                        webcam_pipeline(format!("{}", index).as_str(), fmt),
+                        why.to_string()
+                    ),
+                ))
             }
         };
 
@@ -566,7 +630,8 @@ fn generate_pipeline(fmt: CameraFormat, index: usize) -> Result<PipelineGenRet,
     {
         Some(s) => s,
         None => {
-            return Err(NokhwaError::CouldntOpenDevice(
+            return Err(NokhwaError::OpenDeviceError(
+                index.to_string(),
                 "Failed to get sink element!".to_string(),
             ))
         }
@@ -575,7 +640,8 @@ fn generate_pipeline(fmt: CameraFormat, index: usize) -> Result<PipelineGenRet,
     let appsink = match sink.dynamic_cast::<AppSink>() {
         Ok(aps) => aps,
         Err(_) => {
-            return Err(NokhwaError::CouldntOpenDevice(
+            return Err(NokhwaError::OpenDeviceError(
+                index.to_string(),
                 "Failed to get sink element as appsink".to_string(),
             ))
         }
@@ -643,7 +709,7 @@ fn generate_pipeline(fmt: CameraFormat, index: usize) -> Result<PipelineGenRet,
 
                 let image_buffer = match video_info.format() {
                     VideoFormat::Yuy2 => {
-                        let mut decoded_buffer = match yuyv422_to_rgb888(&buffer_map.as_slice()) {
+                        let mut decoded_buffer = match yuyv422_to_rgb888(&buffer_map) {
                             Ok(buf) => buf,
                             Err(why) => {
                                 element_error!(
@@ -709,7 +775,7 @@ fn generate_pipeline(fmt: CameraFormat, index: usize) -> Result<PipelineGenRet,
                     }
                     // MJPEG
                     VideoFormat::Encoded => {
-                        let mut decoded_buffer = match mjpeg_to_rgb888(&buffer_map.as_slice()) {
+                        let mut decoded_buffer = match mjpeg_to_rgb888(&buffer_map) {
                             Ok(buf) => buf,
                             Err(why) => {
                                 element_error!(
